import sys
from unittest.mock import MagicMock


def test_basic(stub):

    from pytest_stub.utils import Stub

    class Settings(object):
        pass

    modules = sys.modules

    assert 'django.core.management.call_command' not in modules
    assert 'django.core.management.base.BaseCommand' not in modules
    assert 'django.conf.settings' not in modules

    stub.apply({
        'virtualfakedlib': '[mock]',
        'django.faked': '[mock]',
        'django.core.management.call_command': '[func]',
        'django.core.management.base.BaseCommand': '[cls]',
        'django.conf': {
            'settings': Settings(),
            'some': True,
        }
    })
    
    assert isinstance(modules['django.core.management'], Stub)
    assert isinstance(modules['django.core.management.base'], Stub)
    assert isinstance(modules['django.conf'], Stub)

    assert 'django.core.management' in modules
    assert 'django.core.management.base' in modules
    assert 'django.conf' in modules

    from django.conf import settings, some
    from django.core.management import call_command
    from django.core.management.base import BaseCommand

    assert some
    assert isinstance(settings, Settings)
    assert callable(call_command)
    assert isinstance(BaseCommand, object)

    # Test mocking.
    from django import faked
    faked.dummy = 1
    assert faked.dummy == 1

    # Test entire module stub.
    import virtualfakedlib as vl
    some = vl.virtual
    assert isinstance(some, MagicMock)



import pytest

from .utils import StubProxy


@pytest.fixture
def stub():
    """Fixture allowing to temporarily override stub packages, modules and attributes.

    Example::

        def test_django_related(stub):

            stub.apply({
                'cv2': '[mock]',
                'django.dummy': '[mock]',
                'django.core.management.call_command': '[func]',
                'django.core.management.base.BaseCommand': '[cls]',
                'django.conf': {
                    'settings': object(),
                    'some': True,
                }
            })

    :rtype: SettingsProxy

    """
    proxy = StubProxy()

    yield proxy

    proxy.restore_initial()



import sys
from unittest.mock import MagicMock

_UNSET = set()


class StubProxy:

    def __init__(self):
        self._moules = sys.modules
        self._overridden = {}

    def _set(self, path, attrs):

        modules = self._moules
        overridden = self._overridden

        if not isinstance(attrs, dict):  # actual value
            try:
                path, attr = path.rsplit('.', 1)

            except ValueError:
                attr = '__module__'

            attrs = {attr: attrs}

        path = str(path)

        val_current = getattr(modules, path, _UNSET)

        if path not in overridden:
            overridden[path] = val_current

        modules[path] = Stub(path, attrs=attrs or {})

    def apply(self, rules):
        """Apply stubbing rules.

        :param dict rules:

        """
        for path, attrs in rules.items():
            self._set(path, attrs)

    def restore_initial(self):
        """Restore initial modules."""

        modules = self._moules
        unset = _UNSET

        for key, val in self._overridden.items():

            if val is unset:
                # wipe
                del modules[key]

            else:
                # restore
                modules[key] = val


class Stub:

    def __init__(self, path, attrs):
        self.path = path
        self.attrs = attrs

    def __getattr__(self, name):
        attr = self.attrs.get(name, None)

        if attr is None:
            module_mock = self.attrs.get('__module__')

            if module_mock is not None:
                attr = module_mock

            else:
                raise AttributeError('Path %s contain no %s attribute' % (self.path, name))

        if attr == '[cls]':
            return type(str('%sStub' % name), (object,), {})

        elif attr == '[func]':
            return lambda *args, **kwargs: None

        elif attr == '[mock]':
            return MagicMock()

        return attr
